import json

import pytest
from httpsuite import Headers, Request


class Test_request_init_:
    def test_request_init_small(self):
        Request(
            method="GET",
            target="google.com:80",
            protocol="HTTP/1.1",
        )

    def test_request_init_medium(self):
        Request(
            method="GET",
            target="google.com:80",
            protocol="HTTP/1.1",
            headers=None,
            body=None,
        )

    def test_request_init_large(self):
        Request(
            method="GET",
            target="/",
            protocol="HTTP/1.1",
            headers=Headers(
                {
                    "Host": "www.google.com",
                    "Accept": "*/*",
                    "User-Agent": "httpsuite/1.0.0",
                    "Connection": "keep-alive",
                }
            ),
            body=json.dumps({"hello": "world"}),
        )

    def test_init_wrong_type(self):
        with pytest.raises(TypeError):
            Request(method={}, target={}, protocol={})


request = Request(
    method="GET",
    target="/",
    protocol="HTTP/1.1",
    headers=Headers(
        {
            "Host": "www.google.com",
            "Accept": "*/*",
            "User-Agent": "httpsuite/1.0.0",
            "Connection": "keep-alive",
        }
    ),
    body=json.dumps({"hello": "world"}),
)

request_raw = (
    b"POST / HTTP/1.1\r\n"
    b"Host: www.google.com\r\n"
    b"Accept: */*\r\n"
    b"User-Agent: httpsuite/1.0.0\r\n"
    b"Connection: keep-alive\r\n"
    b"\r\n"
    b'{"hello": "world"}'
)


class Test_request_getters:
    def test_request_getter_method(self):
        for method in ["GET", b"GET"]:
            assert request.method == method

    def test_request_getter_target(self):
        for target in ["/", b"/"]:
            assert request.target == target

    def test_request_getter_protocol(self):
        for protocol in ["HTTP/1.1", b"HTTP/1.1"]:
            assert request.protocol == protocol


class Test_request_setters:
    def test_request_setter_method(self):
        request.method = "POST"

        for method in ["POST", b"POST"]:
            assert request.method == method

    def test_request_setter_target(self):
        request.target = "/index.html"

        for target in ["/index.html", b"/index.html"]:
            assert request.target == target

    def test_request_setter_protocol(self):
        request.protocol = "HTTP/0.9"

        for protocol in ["HTTP/0.9", b"HTTP/0.9"]:
            assert request.protocol == protocol


class Test_request_str:
    def test_request_str(self):
        assert "→" in request.__str__()


class Test_request_parse:
    def test_request_str(self):
        parsed = Request.parse(request_raw)
        assert isinstance(parsed, Request)

        assert parsed.method == "POST"
        assert parsed.target == "/"
        assert parsed.protocol == "HTTP/1.1"
        assert parsed.headers == Headers(
            {
                "Host": "www.google.com",
                "Accept": "*/*",
                "User-Agent": "httpsuite/1.0.0",
                "Connection": "keep-alive",
            }
        )
        assert parsed.body == '{"hello": "world"}'



import json

import pytest
from httpsuite import Headers, Response


class Test_response_init_:
    def test_response_init_small(self):
        Response(protocol="HTTP/1.1", status=200, status_msg="OK")

    def test_response_init_large(self):
        Response(
            protocol="GET",
            status=200,
            status_msg="OK",
            headers=Headers(
                {
                    "Host": "www.google.com",
                    "Accept": "*/*",
                    "Connection": "keep-alive",
                }
            ),
            body=json.dumps({"hello": "world"}),
        )

    def test_init_wrong_type(self):
        with pytest.raises(TypeError):
            # pylint: disable=unexpected-keyword-arg, no-value-for-parameter
            Response(method=None, target=None, protocol=None)


response = Response(
    protocol="HTTP/1.1",
    status=200,
    status_msg="OK",
    headers=Headers(
        {
            "Host": "www.google.com",
            "Accept": "*/*",
            "Connection": "keep-alive",
        }
    ),
    body=json.dumps({"hello": "world"}),
)

response_raw = (
    b"HTTP/1.1 200 OK\r\n"
    b"Test: Headers\r\n"
    b"Host: www.google.com\r\n"
    b"Accept: */*\r\n"
    b"Connection: keep-alive\r\n"
    b"\r\n"
    b'{"hello": "world"}'
)


class Test_response_getters:
    def test_response_getter_protocol(self):
        for protocol in ["HTTP/1.1", b"HTTP/1.1"]:
            assert response.protocol == protocol

    def test_response_getter_status(self):
        for status in [200, "200", b"200"]:
            assert response.status == status

    def test_response_getter_status_msg(self):
        for status_msg in ["OK", b"OK"]:
            assert response.status_msg == status_msg


class Test_response_setters:
    def test_response_setter_protocol(self):
        response.protocol = "HTTP/0.9"

        for protocol in ["HTTP/0.9", b"HTTP/0.9"]:
            assert response.protocol == protocol

    def test_response_setter_status(self):
        response.status = "400"

        for status in [400, "400", b"400"]:
            assert response.status == status

    def test_response_setter_status_msg(self):
        response.status_msg = "Not Found"

        for status_msg in ["Not Found", b"Not Found"]:
            assert response.status_msg == status_msg


class Test_response_str:
    def test_request_str(self):
        assert "←" in response.__str__()


class Test_response_parse:
    def test_response_str(self):
        parsed = Response.parse(response_raw)
        assert isinstance(parsed, Response)

        assert parsed.protocol == "HTTP/1.1"
        assert parsed.status == 200
        assert parsed.status_msg == "OK"
        assert parsed.headers == Headers(
            {
                "Test": "Headers",
                "Host": "www.google.com",
                "Accept": "*/*",
                "Connection": "keep-alive",
            }
        )
        assert parsed.body == '{"hello": "world"}'



"""
This test file ensures that toolbox.Item is properly implemented for httpsuite's needs.
Note that the author of httpsuite is the same as toolbox.
"""

from toolbox import Item
import pytest

items = Item("str"), Item(b"bytes"), Item(200)
valid_types = ["str", b"bytes", 200, *[item for item in items]]
invalid_types = [["List"], ("Tuple",), {"Set"}, {"Dictionary": ""}]


class Test_item_init:
    @pytest.mark.parametrize("other", valid_types)
    def test_item_init_valid_types(self, other):
        Item(other)

    @pytest.mark.parametrize("other", invalid_types)
    def test_item_init_invalid_types(self, other):
        with pytest.raises(TypeError):
            Item(other)


class Test_item_string_and_raw:
    @pytest.mark.parametrize("other", items)
    def test_item_string(self, other):
        assert isinstance(Item(other).string, str)
        assert Item(other).string in ("str", "bytes", "200")

    @pytest.mark.parametrize("other", items)
    def test_item_raw(self, other):
        assert isinstance(Item(other).raw, bytes)
        assert Item(other).raw in (b"str", b"bytes", b"200")

    @pytest.mark.parametrize("other", items)
    def test_item_str_and_string(self, other):
        assert str(Item(other)) == Item(other).string


class Test_item_eq:
    def test_item_eq(self):
        assert Item(200) in items
        assert Item("200") in items
        assert Item(b"200") in items


class Test_item_ne:
    def test_item_str_ne(self):
        assert Item(300) not in items
        assert Item("300") not in items
        assert Item(b"300") not in items


class Test_item_add:
    def test_item_add_eq(self):
        assert Item("<html>") + Item("</html>") == Item("<html></html>")
        assert Item("<html>") + "</html>" == Item("<html></html>")
        assert Item("<html>") + b"</html>" == Item("<html></html>")
        assert Item("<html>") + 100 + "</html>" == Item("<html>100</html>")

    def test_item_add_ne(self):
        with pytest.raises(AssertionError):
            assert Item("<html>") + Item("</html>") == Item("<html>")
        with pytest.raises(AssertionError):
            assert Item("<html>") + "<html>" == Item("<html>")
        with pytest.raises(AssertionError):
            assert Item("<html>") + b"<html>" == Item("<html>")
        with pytest.raises(AssertionError):
            assert Item("<html>") + 100 + "</html>" == Item("<html>")


class Test_item_iadd:
    def test_item_iadd_eq(self):
        item = Item("<html>")

        item += Item("</html>")
        assert item == Item("<html></html>")

        item += "</html>"
        assert item == Item("<html></html></html>")

        item += b"</html>"
        assert item == Item("<html></html></html></html>")

        item += 100
        assert item == Item("<html></html></html></html>100")

    def test_item_add_ne(self):
        item = Item("<html>")

        with pytest.raises(AssertionError):
            item += Item("</html>")
            assert item == Item("")

        with pytest.raises(AssertionError):
            item += "<html>"
            assert item == Item("")

        with pytest.raises(AssertionError):
            item += b"<html>"
            assert item == Item("")

        with pytest.raises(AssertionError):
            item += 100
            assert item == Item("")


class Test_item_hash:
    def test_item_hash_eq(self):
        assert hash(Item("<html>")) == hash(Item("<html>"))

    def test_item_hash_ne(self):
        with pytest.raises(AssertionError):
            assert hash(Item("<html>")) == hash(Item("</html>"))



import json

from httpsuite import Request, ENCODE

# 1. Creates the body of the request.
body = json.dumps({"hello": "world"})

# 2. Creates an HTTP request.
request = Request(
    method="GET",
    target="/",
    protocol="HTTP/1.1",
    headers={
        "Host": "www.google.com",
        "Connection": "keep-alive",
        "Content-Length": len(body),
    },
    body=body,
)

# 3. Parses the equivalent request as the above.
request_parsed = request = Request.parse(
    (
        b"GET / HTTP/1.1\r\n"
        b"Host: www.google.com\r\n"
        b"Connection: keep-alive\r\n"
        b"Content-Length: %i\r\n"
        b"\r\n"
        b"%b"
    )
    % (len(body), body.encode(ENCODE))
)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import os
import sys

# Adds system path two folders back (where project lives.)
sys.path.insert(0, os.path.abspath("../.."))

# PyLint might complain, but the interpreter should be able to find this on run.
import httpsuite

# -- Project information -----------------------------------------------------

project = "httpsuite"
copyright = "2020, Felipe Faria"
author = "Felipe Faria"

# -- General configuration ---------------------------------------------------

# Project Name
html_title = "{}".format(project)

# Order of docs.
autodoc_member_order = "bysource"

# Turn off typehints.
autodoc_typehints = "none"

# Remove module names from class docs.
add_module_names = False

# Show only class docs.
autoclass_content = "class"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.viewcode",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "m2r2",
    "sphinx_copybutton",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "furo"

html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



""" Very primative example of a socket microservice architecture using httpsuite.

Server and Client functions are ran through the 'multiprocessing' module, so
to act as two seperate entities. For the sake of clarity, only the server function
prints anything to console. Entities are documented seperately.
"""

from httpsuite import Request, Response
from multiprocessing import Process
import socket
import time


def server():
    """Simple socket server that uses httpsuite to interpret and reply.

    1. Opens a new socket.
    2. Binds to 127.0.0.1:8080 and waits until new connection.
    3. Accepts connection from external source.
    4. Receive the data from the client.
    5. Parse the clients request.
    6. Interpret the request.
    7. Reply to the client.
    8. Close the connection with the client.
    """

    # 1. Opens a new socket.
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    # 2. Binds to 127.0.0.1:8080 and waits until new connection.
    s.bind(("127.0.0.1", 8080))
    s.listen(1)

    # 3. Accepts connection from external source.
    conn, address = s.accept()

    print("===== Connecting With New Client =====", "\n")
    print(address, "\n")

    # 4. Receive the data from the client.
    data = conn.recv(1024)

    # 5. Parse the clients request.
    request = Request.parse(data)

    print("===== Received Data From Client =====", "\n")
    print(request, "\n")

    # 6. Interpret the request.
    response = Response(protocol="HTTP/1.1", status=200, status_msg="OK")
    if request.target == "/":
        response.body = "Homepage of the microservice."
    elif request.target == "/data":
        response.body = "You are accessing the /data directory of this microservice."
    else:
        response.status = 404
        response.status_msg = "Not Found"

    print("===== Replying to Client =====", "\n")
    print(response, "\n")

    # 7. Reply to the client.
    conn.sendall(response.raw)

    print("===== Closing Connection to Client =====", "\n")

    # 8. Close the connection with the client, and the server.
    conn.close()
    s.close()


def client():
    """Simple socket client that uses httpmodule to request server resource.

    1. Opens a new socket.
    2. Connects the server.
    3. Creates a valid request to send to the server.
    4. Sends the request.
    5. Receives reply from the server.
    6. Parses the reply from the server.
    7. Closes connection with the server.
    """

    # Note: Sleeps so that the socket server can boot-up before.
    time.sleep(1)

    # 1. Opens a new socket.
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    # 2. Connects the server.
    s.connect(("127.0.0.1", 8080))

    # 3. Creates a valid request to send to the server.
    request = Request(method="GET", target="/", protocol="HTTP/1.1")

    # 4. Sends the request.
    s.sendall(request.raw)

    # 5. Receives reply from the server.
    data = s.recv(1024)

    # 6. Parses the reply from the server.
    response = Response.parse(data)

    # 7. Closes connection with the server.
    s.close()


if __name__ == "__main__":
    p1 = Process(target=server)
    p2 = Process(target=client)

    p1.daemon = True
    p2.daemon = True

    p1.start()
    p2.start()

    time.sleep(3)
    raise SystemExit



import json

import pytest
from httpsuite import Headers, Request, Response
from toolbox.collections.item import ENCODE

# ====================
#       Request
# ====================
request_headers = {
    "Host": "www.google.com",
    "Accept": "*/*",
    "User-Agent": "httpsuite/1.0.0",
    "Connection": "keep-alive",
}
request_body = json.dumps({"hello": "world"})
request = Request(
    method="GET",
    target="/",
    protocol="HTTP/1.1",
    headers=Headers(request_headers),
    body=request_body,
)
request_raw = b'GET / HTTP/1.1\r\nHost: www.google.com\r\nAccept: */*\r\nUser-Agent: httpsuite/1.0.0\r\nConnection: keep-alive\r\n\r\n{"hello": "world"}'
request_first_line = b"GET / HTTP/1.1"

# ====================
#       Response
# ====================
response_headers = {
    "Host": "www.google.com",
    "Accept": "*/*",
    "Connection": "keep-alive",
    "Keep-Alive": "timeout=5, max=1000",
    "Server": "httpsuite/1.0.0",
}
response_body = json.dumps({"hello": "world"})
response = Response(
    protocol="HTTP/1.1",
    status=200,
    status_msg="OK",
    headers=response_headers,
    body=response_body,
)
response_raw = b'HTTP/1.1 200 OK\r\nHost: www.google.com\r\nAccept: */*\r\nConnection: keep-alive\r\nKeep-Alive: timeout=5, max=1000\r\nServer: httpsuite/1.0.0\r\n\r\n{"hello": "world"}'
response_first_line = b"HTTP/1.1 200 OK"

# ====================
#       Objects
# ====================
objects_headers = [request_headers, response_headers]
objects_body = [request_body, response_body]
objects = [request, response]
objects_raw = [request_raw, response_raw]
objects_first_line = [request_first_line, response_first_line]


class Test_message_getters:
    @pytest.mark.parametrize("zipped", zip(objects, objects_raw))
    def test_message_getter_string(self, zipped):
        assert zipped[0].string == zipped[1].decode(ENCODE)

    @pytest.mark.parametrize("zipped", zip(objects, objects_raw))
    def test_message_getter_raw(self, zipped):
        assert zipped[0].raw == zipped[1]

    @pytest.mark.parametrize("zipped", zip(objects, objects_first_line))
    def test_message_getter_first_line(self, zipped):
        assert zipped[0].first_line == zipped[1]

    @pytest.mark.parametrize("zipped", zip(objects, objects_headers))
    def test_message_getter_headers_valid(self, zipped):
        assert zipped[0].headers == zipped[1]
        assert zipped[0].headers == Headers(zipped[1])

    @pytest.mark.parametrize("zipped", zip(objects, objects_body))
    def test_message_getter_body(self, zipped):
        assert zipped[0].body == zipped[1]
        assert zipped[0].body == bytes(zipped[1], ENCODE)

    @pytest.mark.parametrize("message", objects)
    def test_message_getter_protocol(self, message):
        assert message.protocol == "HTTP/1.1"
        assert message.protocol == b"HTTP/1.1"


class Test_message_setters:
    def test_message_setter_request_headers_valid(self):
        request.headers = {}
        request.headers = Headers({})

    def test_message_setter_response_headers_valid(self):
        response.headers = {}
        response.headers = Headers({})

    def test_message_setter_request_headers_invalid(self):
        with pytest.raises(TypeError):
            request.headers = "NONE"

    def test_message_setter_response_headers_invalid(self):
        with pytest.raises(TypeError):
            response.headers = "NONE"

    def test_message_setter_request_body(self):
        request.body = "<html>"

    def test_message_setter_response_body(self):
        response.body = "<html>"


class Test_message_compile:
    def test_message_compile_response(self):
        with pytest.raises(TypeError):
            assert response._compile(frmt="dict")

    def test_message_compile_request(self):
        with pytest.raises(TypeError):
            assert request._compile(frmt="dict")



# -*- coding: utf-8 -*-
""" Classes for parsing, modifying, and re-compiling HTTP messages. """

import abc
from typing import Any, Type, Union

from toolbox.collections.item import Item, ItemType, ENCODE

HeadersType = Union[dict, "Headers", None]


class Headers(dict):
    """Interface for HTTP/1.x headers object."""

    def __init__(self, headers: HeadersType = None) -> None:
        """Initializes the Headers object.

        Args:
            value: Dictionary, headers, or None object that represents the headers.
        """

        if headers is None:
            headers = {}

        self._check(headers)

        if isinstance(headers, dict):
            for key, value in headers.items():
                self[Item(key)] = Item(value)

        super().__init__()

    def _compile(self, frmt: str = "bytes") -> Union[str, bytes]:
        r"""Compiles the Headers into the passed format.

        Notes:
            When the format is bytes this function will return the headers in bytes
            format with the correct HTTP escape characters "\r\n". This does **not**
            occur when string is passed as the format (new lines are created instead).

        Args:
            format: Either bytes or str. Formats the return accordingly to the passed
                format.

        Returns:
            String or bytes representation of the Headers.
        """

        if frmt == "bytes":
            data = b""
            for key, value in self.items():
                data += b"%b: %b\r\n" % (key.raw, value.raw)

        elif frmt == "string":
            data = ""
            for key, value in self.items():
                data += "{}: {}\r\n".format(key.string, value.string)

            if data[-2:] == "\r\n":
                data = data[: len(data) - 2]

        return data

    def _check(self, value: HeadersType) -> Union[None, None]:
        """Checks the item type to verify it is either Headers or dict.

        Raises:
            TypeError: Headers is not either dict or headers.
        """

        if not isinstance(value, dict) and not isinstance(value, Headers):
            raise TypeError("Headers can only be of type that inherits from 'dict'.")

    @property
    def string(self) -> str:
        """String representation of the Headers.

        Returns:
            String representation of the Headers.
        """
        return self._compile(frmt="string")

    @property
    def raw(self) -> bytes:
        r"""Bytes representation of the Headers.

        Note:
            This method will return Headers with "\r\n" escape characters.

        Returns:
            Bytes representation of the Headers.
        """
        return self._compile(frmt="bytes")

    def __add__(self, other: Union[dict, "Headers"]) -> "Headers":
        """Adds item with passed other and returns new Headers.

        Args:
            other: Headers or dictionary to be added.

        Returns:
            Headers: New resulting Headers object.
        """

        self._check(other)

        if isinstance(other, Headers):
            current = self.copy()
            current.update(other)
            return Headers(current)
        elif isinstance(other, dict):
            itemized_values = {Item(k): Item(v) for k, v in other.items()}
            copy = self.copy()
            copy.update(itemized_values)
            return copy

    def __iadd__(self, other: Union[dict, "Headers"]) -> "Headers":
        """Adds current headers with passed other and returns self.

        Args:
            other: Headers or dictionary to be added.

        Returns:
            Headers: Self after addition of other to itself.
        """

        self._check(other)

        itemized_values = {Item(k): Item(v) for k, v in other.items()}
        self.update(itemized_values)
        return self

    def __setattr__(self, key: str, value: str) -> None:
        """Sets a new attribute inside Headers.

        Notes:
            If "_" is present in the key it gets replaced with "-". This is so, example,
            "headers.User_Agent" is equivalent to "headers['User-Agent']".
        """

        key_mod = key.replace("_", "-").encode(ENCODE)
        self[Item(key_mod)] = Item(value)

    def __getattr__(self, key: str) -> Item:
        """Gets attribute inside Headers.

        Notes:
            If "_" is present in the key it gets replaced with "-". This is so the
            "headers.User_Agent" is equivalent to "headers['User-Agent']".

        Returns:
            Item corresponding to the passed key.
        """

        key_mod = key.replace("_", "-").encode(ENCODE)

        if key_mod in self:
            return self[key_mod]
        else:
            return None

    def __str__(self) -> str:
        """String representation of the Headers.

        Returns:
            String representation of the Headers.
        """
        return self.string

    def __repr__(self):
        """String representation of the Headers.

        Returns:
            String representation of the Headers.
        """
        return "Headers({})".format(self.string)


class Message(abc.ABC):
    """Abstract class that contains shared methods and properties accessible by both the
    Request and Response classes.

    Warning:
        This class is not intended to be used by itself. Note that Message is an
        abstraction that represents the shared properties and methods of both a Request
        and Response instance. All functions displayed in Message are, therefore,
        accessible by both the Request and Response classes.
    """

    __slots__ = ["_protocol", "_headers", "_body", "_first_line"]

    def __init__(
        self, protocol: ItemType, headers: HeadersType = None, body: ItemType = None
    ) -> None:
        """Initializes an HTTP Message.

        Args:
            protocol: "<major>.<minor>" numbering scheme to indicate versions of the
                protocol.
            headers: Collection of case-insensitive name followed by a colon (:).
            body: Data associated with the message.
        """
        self._first_line = None
        self._protocol = Item(protocol)

        if isinstance(headers, Headers):
            self._headers = headers
        elif headers is None:
            self._headers = Headers()
        else:
            self._headers = Headers(headers)

        self._body = Item(body)

    # ================ Parsing ================
    #           other types -> Message

    @classmethod
    def parse(cls: Union["Message", "Response", "Request"], message: Any) -> None:
        """Parses a raw HTTP message (in bytes or string) to an object instance of class 'cls'.

        Args:
            message: The primative or object to convert into a Request or Response
                object.

        Returns:
            An initialized object of class 'cls'.
        """

        if isinstance(cls, Message):
            err = "You can only use .parse() with either the Request or Response class."
            raise ValueError(err)

        if not isinstance(message, Item):
            message = Item(message)

        first_line = b""
        headers = Headers({})
        body = b""

        top_frame = True
        for index, line in enumerate(message.raw.splitlines()):
            if line == b"":
                top_frame = False

            if top_frame:
                if index == 0:
                    first_line = line
                elif b":" in line:
                    key, value = line.split(b":", 1)
                    headers += {key: value.lstrip(b" ")}
            else:
                body += line

        args = (*first_line.split(b" "), headers, body)

        try:
            return cls(*args)
        except TypeError as error:  # pragma: no cover
            err = "Error parsing the message due to {}.".format(error)
            raise TypeError(err) from error

    # ================ Compilation ================
    #             Message -> other types

    def _compile(
        self, frmt: type = Union[Type[str], Type[bytes]], arrow: str = ""
    ) -> Union[str, bytes]:
        """Compile the Message object into a useable type (str, or bytes).

        Note:
            The 'arrow' argument only works when 'frmt' is 'str'.

        Args:
            format: Type that represents the return type. Either 'str' or 'bytes'.
            arrow: String to append to the beginning of every line.

        Returns:
            String or bytes representation of the 'Message'.
        """

        if frmt is not bytes and frmt is not str:
            raise TypeError("format must either be str, or byte.")

        self._compile_first_line()

        if frmt is bytes:
            first_line = self._first_line.raw
            body = self._body.raw
            headers = b""
        elif frmt is str:
            first_line = self._first_line.string
            body = self._body.string
            headers = ""

        for key, value in self.headers.items():
            if frmt is bytes:
                headers += b"%b: %b\r\n" % (key.raw, value.raw)
            elif frmt is str:
                headers += "{}: {}\r\n".format(key.string, value.string)

        if frmt is bytes:
            msg = b"%b\r\n%b\r\n%b" % (first_line, headers, body)
        elif frmt is str:
            msg = "{}\r\n{}\r\n{}".format(first_line, headers, body)

        if frmt is str and arrow:
            arrow_msg = [
                "{} {}".format(arrow, line) for line in msg.splitlines() if line
            ]
            return "\r\n".join(arrow_msg)
        else:
            return msg

    # ----- Conversions -----

    @abc.abstractmethod
    def _compile_first_line(self) -> None:  # pragma: no cover
        """Compiles the first line of the message.

        Notes:
            Sets self._first_line to self.protocol; self.method or self.status;
            self.target or self.status_msg.
        """
        raise NotImplementedError

    # ================ Properties ================

    # Protocol, headers, and body are the three
    # shared properties between an HTTP request, and
    # response.

    @property
    def protocol(self) -> Item:
        """Protocol of the message."""
        return self._protocol

    @protocol.setter
    def protocol(self, value: ItemType) -> None:
        self._protocol = Item(value)

    @property
    def headers(self) -> Headers:
        """Headers of the message."""
        return self._headers

    @headers.setter
    def headers(self, value: Union[dict, Headers, None]) -> None:
        if not isinstance(value, dict):
            raise TypeError("can only set to type that inherits from 'dict'.")

        if isinstance(value, Headers):
            self._headers = value
        else:
            self._headers = Headers(value)

    @property
    def body(self) -> Item:
        """Body of the message."""
        return self._body

    @body.setter
    def body(self, value: ItemType) -> None:
        self._body = Item(value)

    @property
    def first_line(self) -> Item:
        """First line of the Message.

        Returns:
            The first line of the Message.
        """

        self._compile_first_line()
        return self._first_line

    @property
    def __items__(self) -> dict:
        """Fast retrieval of the Message's content.

        Returns:
            Dictionary with the message's protocol, headers, and body.
        """
        return {"protocol": self.protocol, "headers": self.headers, "body": self.body}

    # ----- Compiled Properties -----

    @property
    def string(self) -> str:
        """String representation of the Message.

        Returns:
            Message as a string, without any arrows.
        """

        return self._compile(frmt=str)

    @property
    def raw(self) -> bytes:
        """Bytes representation of the Message.

        Returns:
            Message as a bytes, without arrows, properly escaped.
        """
        return self._compile(frmt=bytes)


class Request(Message):
    __slots__ = ["_method", "_target"]

    def __init__(
        self,
        method: ItemType,
        target: ItemType,
        protocol: HeadersType,
        headers: ItemType = None,
        body: ItemType = None,
    ) -> None:
        """Python object representation of an HTTP/1.x request."

        Args:
            method: Indicates the desired action on the server's resource.
            target: Resource location in the server for which the client is requesting.
            protocol: "<major>.<minor>" numbering scheme to indicate versions of the
                protocol.
            headers: Collection of case-insensitive name followed by a colon (:).
            body: Data associated with the message.
        """

        self._method = Item(method)
        self._target = Item(target)

        super().__init__(protocol, headers, body)

        first_line = (self._method.raw, self._target.raw, self._protocol.raw)
        self._first_line = Item(b"%b %b %b" % first_line)

    # ================ Compilation ================

    def _compile_first_line(self) -> None:
        """Compiles the first line of the message.

        Notes:
            Sets self._first_line to self.protocol, self.method, self.target.
        """
        first_line = (self._method.raw, self._target.raw, self._protocol.raw)
        self._first_line = Item(b"%b %b %b" % first_line)

    # ================ Properties ================

    @property
    def method(self) -> Item:
        """Method of the HTTP request."""
        return self._method

    @method.setter
    def method(self, value: ItemType) -> None:
        self._method = Item(value)

    @property
    def target(self) -> Item:
        """Target of the HTTP request."""
        return self._target

    @target.setter
    def target(self, value: ItemType) -> None:
        self._target = Item(value)

    @property
    def __items__(self) -> dict:
        """Items corresponding to the Request."""
        return {
            **super().__items__,
            **{
                "method": self.method,
                "target": self.target,
            },
        }

    # ================ Misc ================

    def __str__(self) -> str:
        """String representation of the Request.

        Returns:
            Representation of the Request object with pretty-print (→) arrows.
        """
        return self._compile(frmt=str, arrow="→")


class Response(Message):
    __slots__ = ["_status", "_status_msg"]

    def __init__(
        self,
        protocol: ItemType,
        status: ItemType,
        status_msg: ItemType,
        headers: HeadersType = None,
        body: ItemType = None,
    ) -> None:
        """Python object representation of an HTTP/1.x response.

        Args:
            protocol: "<major>.<minor>" numbering scheme to indicate versions of the
                protocol.
            status: Numberical value designating a specific return value.
            status_msg: Message related to the status code.
            headers: Collection of case-insensitive name followed by a colon (:).
            body: Data associated with the message.
        """

        self._status = Item(status)
        self._status_msg = Item(status_msg)

        super().__init__(protocol, headers, body)

        first_line = self._protocol.raw, self._status.raw, self._status_msg.raw
        self._first_line = Item(b"%b %b %b" % first_line)

    # ================ Compilation ================

    def _compile_first_line(self) -> None:
        """Compiles the first line of the message.

        Notes:
            Sets self._first_line to self.protocol, self.status, self.status_msg.
        """
        first_line = self._protocol.raw, self._status.raw, self._status_msg.raw
        self._first_line = Item(b"%b %b %b" % first_line)

    # ================ Properties ================

    @property
    def status(self) -> Item:
        """Status of the HTTP response."""
        return self._status

    @status.setter
    def status(self, value: ItemType) -> None:
        self._status = Item(value)

    @property
    def status_msg(self) -> Item:
        """Status message of the HTTP response."""
        return self._status_msg

    @status_msg.setter
    def status_msg(self, value: ItemType) -> None:
        self._status_msg = Item(value)

    @property
    def __items__(self) -> dict:
        """Items corresponding to the Response."""

        return {
            **super().__items__,
            **{
                "status": self.status,
                "status_msg": self.status_msg,
            },
        }

    # ================ Misc ================

    def __str__(self) -> str:
        """String representation of the Response.

        Returns:
            Representation of the Response object with pretty-print (←) arrows.
        """
        return self._compile(frmt=str, arrow="←")



""" Example of using httpsuite to communicate with an external web server using sockets.

1. Open a new socket.
2. Creates the request to be sent via Request object.
3. Connect to httpbin.org via socket.
4. Send generated request to server.
5. Receive raw response from server.
6. Parse the server's response with Response object.
7. Loads the response's body via JSON.
8. Close socket.
"""

from httpsuite import Request, Response
import socket
import json

# 1. Open a new socket.
s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

# 2. Creates the request to be sent via Request object.
body = json.dumps({"hello": "world"})
request = Request(
    method="POST",
    target="/post",
    protocol="HTTP/1.1",
    headers={
        "Host": "httpbin.org",
        "Connection": "close",
        "Content-Length": len(body),
        "Accept": "*/*",
    },
    body=body,
)

# Prints the raw request.
print("====== Raw Request ======", "\n")
print(request.raw, "\n")

# 3. Connect to httpbin.org via socket.
s.connect(("httpbin.org", 80))

# 4. Send generated request to server.
s.sendall(request.raw)

# 5. Receive raw response from server.
response_raw = s.recv(4096)

# Prints the raw response.
print("====== Raw Response ======", "\n")
print(response_raw, "\n")

# 6. Parse the server's response with Response object.
response = Response.parse(response_raw)

# Prints the request and the response (pretty-print).
print("====== Request and Response ======", "\n")
print(request, "\n")
print(response, "\n")

# 7. Loads the response's body via JSON.
body = json.loads(response.body.string)

# Prints the loaded json ('dumps' for pretty-print).
print("====== Json ======", "\n")
print(json.dumps(body, indent=4))

# 8. Close socket.
s.close()



from httpsuite import Headers
import pytest

dictionaries = [{"str": "str"}, {b"bytes": b"bytes"}]
headers = [Headers(d.copy()) for d in dictionaries]
valid_types = [{"str": "str"}, {b"bytes": b"bytes"}, *[header for header in headers]]
invalid_types = [
    "str",
    b"bytes",
    200,
    ["List"],
    ("Tuple",),
    {"Set"},
]


class Test_headers_init:
    @pytest.mark.parametrize("other", valid_types)
    def test_headers_init_valid_types(self, other):
        Headers(other)

    @pytest.mark.parametrize("other", invalid_types)
    def test_headers_init_invalid_types(self, other):
        with pytest.raises(TypeError):
            Headers(other)


class Test_headers_compile_string_raw:
    @pytest.mark.parametrize("other", headers)
    def test_headers_compile_string(self, other):
        compiled = other._compile(frmt="string")
        assert isinstance(compiled, str)
        assert compiled in ("str: str", "bytes: bytes")
        assert other.string == compiled

    @pytest.mark.parametrize("other", headers)
    def test_headers_compile_bytes(self, other):
        compiled = other._compile(frmt="bytes")
        assert isinstance(compiled, bytes)
        assert compiled in (b"str: str\r\n", b"bytes: bytes\r\n")
        assert other.raw == compiled


class Test_headers_add:
    def test_headers_add_headers(self):
        added = headers[0] + headers[1]
        assert added == Headers({"str": "str", "bytes": "bytes"})
        assert added == Headers({"str": "str", b"bytes": b"bytes"})
        assert added == Headers({b"str": b"str", "bytes": b"bytes"})
        assert added == Headers({b"str": b"str", b"bytes": b"bytes"})

    def test_headers_add_headers_uniqueness(self):
        added = headers[0] + headers[1]
        assert id(headers[0]) != id(added)
        assert id(headers[1]) != id(added)

    def test_headers_add_dict(self):
        added = headers[0] + dictionaries[1]
        assert added == Headers({"str": "str", "bytes": "bytes"})
        assert added == Headers({"str": "str", b"bytes": b"bytes"})
        assert added == Headers({b"str": b"str", "bytes": b"bytes"})
        assert added == Headers({b"str": b"str", b"bytes": b"bytes"})

    def test_headers_add_dict_uniqueness(self):
        added = headers[0] + dictionaries[1]
        assert id(headers[0]) != id(added)
        assert id(dictionaries[1]) != id(added)

    @pytest.mark.parametrize("other", invalid_types)
    def test_headers_add_invalid_type(self, other):
        added = Headers()
        with pytest.raises(TypeError):
            added + other


class Test_headers_iadd:
    def test_headers_iadd_headers(self):
        added = Headers()

        for header in headers:
            added += header

        assert added == Headers({"str": "str", "bytes": "bytes"})
        assert added == Headers({"str": "str", b"bytes": b"bytes"})
        assert added == Headers({b"str": b"str", "bytes": b"bytes"})
        assert added == Headers({b"str": b"str", b"bytes": b"bytes"})

    def test_headers_iadd_dict(self):
        added = Headers()

        for dictionary in dictionaries:
            added += dictionary

        assert added == Headers({"str": "str", "bytes": "bytes"})
        assert added == Headers({"str": "str", b"bytes": b"bytes"})
        assert added == Headers({b"str": b"str", "bytes": b"bytes"})
        assert added == Headers({b"str": b"str", b"bytes": b"bytes"})

    @pytest.mark.parametrize("other", invalid_types)
    def test_headers_iadd_invalid_types(self, other):
        added = Headers()
        with pytest.raises(TypeError):
            added += other


class Test_headers_setattr_and_getattr:
    def test_headers_setattr_and_getattr_single_word(self):
        headers = Headers()
        headers.Host = "github.com"

        assert headers.Host == "github.com"
        assert headers.Host == b"github.com"

        assert headers["Host"] == "github.com"
        assert headers["Host"] == b"github.com"
        assert headers[b"Host"] == "github.com"
        assert headers[b"Host"] == b"github.com"

    def test_headers_setattr_and_getattr_double_word(self):
        headers = Headers()
        headers.Cache_Control = "no-cache"

        assert headers.Cache_Control == "no-cache"
        assert headers.Cache_Control == b"no-cache"

        assert headers["Cache-Control"] == "no-cache"
        assert headers["Cache-Control"] == b"no-cache"
        assert headers[b"Cache-Control"] == "no-cache"
        assert headers[b"Cache-Control"] == b"no-cache"


class Test_headers_str:
    def test_headers_str(self):
        for header in headers:
            assert str(header) in ("str: str", "bytes: bytes")

    def test_headers_str_compiled(self):
        for header in headers:
            assert str(header) == header._compile(frmt="string")



from httpsuite.interface import FrozenSet, TwoWayFrozenDict
from toolbox import Item

status = TwoWayFrozenDict({100: "Continue"})
protocols = FrozenSet({"GET"})


class Test_misc_TwoWayFrozenDict:
    def test_misc_TwoWayFrozenDict_init(self):
        assert status

    def test_misc_TwoWayFrozenDict_get_attribute(self):
        assert status._100 == "Continue"
        assert status._100 == b"Continue"

        assert status[100] == "Continue"
        assert status[100] == b"Continue"
        assert status["100"] == "Continue"
        assert status["100"] == b"Continue"
        assert status[b"100"] == "Continue"
        assert status[b"100"] == b"Continue"

        assert status.Continue == 100
        assert status.Continue == "100"
        assert status.Continue == b"100"

        assert status["Continue"] == 100
        assert status[b"Continue"] == 100
        assert status["Continue"] == "100"
        assert status[b"Continue"] == "100"
        assert status["Continue"] == b"100"
        assert status[b"Continue"] == b"100"

    def test_misc_TwoWayFrozenDict_iter(self):
        for item in status:
            assert isinstance(item, Item)

    def test_misc_TwoWayFrozenDict_len(self):
        assert len(status) == status.__len__()

    def test_misc_TwoWayFrozenDict_get_item_in(self):
        assert "100" in status
        assert b"100" in status
        assert 100 in status

        assert "Continue" in status
        assert b"Continue" in status

    def test_misc_TwoWayFrozenDict_get_not_in(self):
        assert "NONE" not in status
        assert not status.__contains__("NONE")


class Test_misc_FrozenSet:
    def test_misc_FrozenSet_init(self):
        assert protocols

    def test_misc_FrozenSet_str(self):
        protocols_str = protocols.__str__()
        assert "<" not in protocols_str and ">" not in protocols_str



# -*- coding: utf-8 -*-
""" Collection of RFC specifications related to HTTP requests and responses.

Every item in this file is commented with it's specific specification, chapter, and section number.
To access these specifications, utilize the URL https://tools.ietf.org/html/<id>.
"""

from .interface import FrozenSet, TwoWayFrozenDict

# Core Rules
# rfc5234#appendix-B.1
CR = b"\r"
LF = b"\n"


# Protocols
PROTOCOLS = FrozenSet(
    {
        "HTTP/0.9",  # rfc1945#section-3.1
        "HTTP/1.0",  # rfc1945#section-3.1
        "HTTP/1.1",  # rfc7231
        "HTTP/2.0",  # rfc7540
        "HTTP/3.0",  # draft-ietf-quic-http-27
    }
)


# Request Method Definitions
# rfc7231#section-4
REQUEST_METHODS = FrozenSet(
    {
        "GET",  # rfc7231#section-4.3.1
        "HEAD",  # rfc7231#section-4.3.2
        "POST",  # rfc7231#section-4.3.3
        "PUT",  # rfc7231#section-4.3.4
        "DELETE",  # rfc7231#section-4.3.5
        "CONNECT",  # rfc7231#section-4.3.6
        "OPTIONS",  # rfc7231#section-4.3.7
        "TRACE",  # rfc7231#section-4.3.8
    }
)

# Request Header Fields
# rfc7231#section-5
REQUEST_HEADERS = FrozenSet(
    {
        # Control
        # rfc7231#section-4
        "Cache-Control",  # rfc7234#section-5.2
        "Except",  # rfc7231#section-5.1.1
        "Host",  # rfc7230#section-5.4
        "Max-Forwards",  # rfc7231#section-5.1.2
        "Pragma",  # rfc7234#section-5.4
        "Range",  # rfc7233#section-3.1"
        "TE",  # rfc7230#section-4.3
        # Conditionals
        # rfc7231#section-5.2
        "If-Match",  # rfc7232#section-3.1
        "If-None-Match",  # rfc7232#section-3.2
        "If-Modified-Since",  # rfc7232#section-3.3
        "If-Unmodified-Since",  # rfc7232#section-3.4
        "If-Range",  # rfc7233#section-3.2
        # Content Negotiation
        # rfc7231#section-5.3
        "Accept",  # rfc7231#section-5.3.2
        "Accept-Charset",  # rfc7231#section-5.3.3
        "Accept-Encoding",  # rfc7231#section-5.3.4
        "Accept-Language",  # rfc7231#section-5.3.5
        # Authentication Credentials
        # rfc7231#section-5.4
        "Authorization",
        "Proxy-Authorization",
        # Request Context
        # rfc7231#section-5.5
        "From",  # rfc7231#section-5.5.1
        "Referer",  # rfc7231#section-5.5.2
        "User-Agent",  # rfc7231#section-5.5.3
    }
)

# Response Status Code
# rfc7231#section-6
RESPONSE_STATUS = TwoWayFrozenDict(
    {
        # Informational 1xx
        # rfc7231#section-6.2
        100: "Continue",  # rfc7231#section-6.2.1
        101: "Switching Protocols",  # rfc7231#section-6.2.2
        # Successful 2xx
        # rfc7231#section-6.3
        200: "OK",  # rfc7231#section-6.3.1
        201: "Created",  # rfc7231#section-6.3.2
        202: "Accepted",  # rfc7231#section-6.3.3
        203: "Non-Authoritative Information",  # rfc7231#section-6.3.4
        204: "No Content",  # rfc7231#section-6.3.5
        205: "Reset Content",  # rfc7231#section-6.3.6
        206: "Partial Content",  # rfc7233#section-4.1
        # Redirection 3xx
        # rfc7231#section-6.4
        300: "Multiple Choices",  # rfc7231#section-6.4.1
        301: "Moved Permanently",  # rfc7231#section-6.4.2
        302: "Found",  # rfc7231#section-6.4.3
        303: "See Other",  # rfc7231#section-6.4.4
        304: "Not Modified",  # rfc7232#section-4.1
        305: "Use Proxy",  # rfc7231#section-6.4.5
        307: "Temporary Redirect",  # rfc7231#section-6.4.7
        # Client Error 4xx
        # rfc7231#section-6.5
        400: "Bad Request",  # rfc7231#section-6.5.1
        401: "Unauthorized",  # rfc7235#section-3.1
        402: "Payment Required",  # rfc7231#section-6.5.2
        403: "Forbidden",  # rfc7231#section-6.5.3
        404: "Not Found",  # rfc7231#section-6.5.4
        405: "Method Not Allowed",  # rfc7231#section-6.5.5
        406: "Not Acceptable",  # rfc7231#section-6.5.6
        407: "Proxy Authentication Required",  # rfc7235#section-3.2
        408: "Request Timeout",  # rfc7231#section-6.5.7
        409: "Conflict",  # rfc7231#section-6.5.8
        410: "Gone",  # rfc7231#section-6.5.9
        411: "Length Required",  # rfc7231#section-6.5.10
        412: "Precondition Failed",  # rfc7232#section-4.2
        413: "Payload Too Large",  # rfc7231#section-6.5.11
        414: "URI Too Long",  # rfc7231#section-6.5.12
        415: "Unsupported Media Type",  # rfc7231#section-6.5.13
        416: "Range Not Satisfiable",  # rfc7233#section-4.4
        417: "Expectation Failed",  # rfc7231#section-6.5.14
        426: "Upgrade Required",  # rfc7231#section-6.5.15
        # Server Error 5xx
        # rfc7231#section-6.6
        500: "Internal Server Error",  # rfc7231#section-6.6.1
        501: "Not Implemented",  # rfc7231#section-6.6.2
        502: "Bad Gateway",  # rfc7231#section-6.6.3
        503: "Service Unavailable",  # rfc7231#section-6.6.4
        504: "Gateway Timeout",  # rfc7231#section-6.6.5
        505: "HTTP Version Not Supported",  # rfc7231#section-6.6.6
    }
)

# Response Header Fields
# rfc7231#section-7
RESPONSE_HEADER = FrozenSet(
    {
        # Control Data
        # rfc7231#section-7.1
        "Age",  # rfc7234#section-5.1
        "Cache-Control",  # rfc7234#section-5.2
        "Expires",  # rfc7234#section-5.3
        "Date",  # rfc7231#section-7.1.1.2
        "Location",  # rfc7231#section-7.1.2
        "Retry-After",  # rfc7231#section-7.1.3
        "Vary",  # rfc7231#section-7.1.4
        "Warning",  # rfc7234#section-5.5
        # Validator Header Fields
        # rfc7231#section-7.2
        "ETag",  # rfc7232#section-2.3
        "Last-Modified",  # rfc7232#section-2.2
        # Authentication Challenges
        # rfc7231#section-7.3
        "WWW-Authenticate",  # rfc7235#section-4.1
        "Proxy-Authenticate",  # rfc7235#section-4.3
        # Response Context
        # rfc7231#section-7.4
        "Accept-Ranges",  # rfc7233#section-2.3
        "Allow",  # rfc7231#section-7.4.1
        "Server",  # rfc7231#section-7.4.2
    }
)



# -*- coding: utf-8 -*-
""" Lower-level interfaces that ``httpsuite`` depedents on. """

from toolbox.collections.mapping import (
    BidirectionalDict,
    FrozenDict,
    ItemDict,
    ObjectDict,
    UnderscoreAccessDict,
)


class TwoWayFrozenDict(
    BidirectionalDict,
    FrozenDict,
    ItemDict,
    ObjectDict,
    UnderscoreAccessDict,
):
    """A frozen dictionary with two-way capabilities. Locks a dictionary in place after
    initilization, and provides accessability via key and value.

    Note:
        All the keys and values inside TwoWayFrozenDict are Item objects, which allows
        easy comparissions to check if an item is inside the TwoWayFrozenDict mapping.
    """


class FrozenSet(frozenset):
    """A frozen set with pretty-print."""

    def __str__(self):
        """String representation of the FrozenSet.

        Returns:
            String representation of the FrozenSet.
        """
        return str({k for k in self})
